
rm(list = ls())

##### Compare 4 methods (DGL,IGL,tMVN,bsar) for non-decreasing functions #####

library(tmg)
library(FastGP) 
library(bsamGP)

### Non-decreasing functions that have a flat region (f1,f2,f3), and strictly increasing function (f4) 
f1 =  function(x) (5 * x-3)^3 * (5 * x-3>0) 
   
f2 = function(x) 3/(1+exp(-10*x+2.1))

f3 = function(x){
    ind = seq(1, 100,1)
    eigind = ind^(-1.7)*sin(ind)
    basind= sqrt(2)*cos(pi*(ind - 0.5)*(1-x))
    val = sum(eigind*basind)
  } 
  
f4 = function(x) 5*x^2   


#### generating the synthetic data ####
set.seed(12345)
n=500
x=sort(runif(n,min=0,max=1))
y.true=sapply(x,FUN = f1); func_ind=1  ## generate response observations from function f1/f2/f3/f4 
sigmaT=0.5
noise=rnorm(n,sd=sigmaT)
y=y.true+noise      # Y-data
print(sd(y.true)/sd(noise))
data=data.frame(x,y.true,y)

## Random split of the data set in to training and testing data sets 
ntr=300; nte=n-ntr
ind=sort(sample(1:n,ntr))
xtrain=x[ind]
ytrain=y[ind]
ytr.tr=y.true[ind]
data_train=data.frame(xtrain,ytr.tr,ytrain)  # training data set
xtest=x[-ind]
ytest=y[-ind]
yte.tr=y.true[-ind]
data_test=data.frame(xtest,yte.tr,ytest)    # testing data set

# determine the flat region and non-flat region for function f1/f2/f3/f4
if(func_ind == 2 || func_ind == 3){
  reg_i = function(x) (x<0.7)
  reg_f = function(x) (x>=0.7)
}else if(func_ind == 1){
  reg_i = function(x) (x>0.6)
  reg_f = function(x) (x<=0.6)
}else if(func_ind == 4){
  reg_i = function(x) (x<=1)
}


###########################################
####  Fit models with compared priors  ####
###########################################

source("~/Dropbox/code-DGL/all_base_functions.R")
source("~/Dropbox/code-DGL/all_models.R")
require(Rcpp)
#sourceCpp("~/Dropbox/code-DGL/rcpptry2.cpp") 
sourceCpp("~/Dropbox/code-DGL/rcpptry_cpp.cpp") # alternative to "rcpptry2.cpp"


##### knots, test design matrix, MCMC parameters, etc. #####
# knots and design matrix:
N=ceiling(ntr/8)-1
int_length=1/N
my_knots=seq(0,1,by=int_length)
# test design matrix \psi(nte X N+1) 
Xte=matrix(0,nte,N+1)
for(i in 1:nte){
  Xte[i,1:(N+1)]=psi_j(xtest[i],my_knots,int_length)
}

# parameter of the approximation function:
eta=50
# MCMC samples:
mcmc=10000
brn=5000
thin=1
em=mcmc+brn
ef=mcmc/thin


##### Model with IGL prior #####
system.time({
  post.sam=mon.horseshoe.hyp(ytrain,xtrain,N=N,eta=eta,sseed=123,mcmc=mcmc,brn=brn,thin=thin,
                           return.plot = FALSE,return.traplot = FALSE)
})
# Prediction
xi_sam=post.sam$xi_sam
xi0_sam=post.sam$xi0_sam
fhat_sam=t(xi0_sam + t(Xte %*% xi_sam))
fmean=rowMeans(fhat_sam)
fsort=fhat_sam
for(i in 1:nrow(fhat_sam)){
  fsort[i,]=sort(fhat_sam[i,])
}
f_low=fsort[,ef*0.025]
f_upp=fsort[,ef*0.975]
# save the posterior samples and predictive values
out = list("fhat_sam"=fhat_sam,"fmean"=fmean,"f_low"=f_low,"f_upp"=f_upp)


##### Model with DGL prior (with Matern kernel, non-decreasing) #####
system.time({
  post.sam=mon.shrink.hyp(ytrain,xtrain,N=N,eta=eta,sseed=123,mcmc=mcmc,brn=brn,thin=thin,
                          return.plot = FALSE,return.traplot = FALSE)
})
# Prediction
xi_sam=post.sam$xi_sam
lam_sam=post.sam$lam_sam
xi0_sam=post.sam$xi0_sam
fhat_sam=t(xi0_sam + t(Xte %*% (xi_sam*lam_sam)))
fmean=rowMeans(fhat_sam)
fsort=fhat_sam
for(i in 1:nrow(fhat_sam)){
  fsort[i,]=sort(fhat_sam[i,])
}
f_low=fsort[,ef*0.025]
f_upp=fsort[,ef*0.975]
# save the posterior samples and predictive values
out = list("fhat_sam"=fhat_sam,"fmean"=fmean,"f_low"=f_low,"f_upp"=f_upp)


##### Model with tMVN prior (with improper prior on \tau^2, non-decreasing) #####
system.time({
  post.sam=mon.ESS.hyp(ytrain,xtrain,N=N,eta=eta,sseed=123,mcmc=mcmc,brn=brn,thin=thin,
                       return.plot = FALSE,return.traplot = FALSE)
})
# Prediction
xi_sam=post.sam$xi_sam
xi0_sam=post.sam$xi0_sam
fhat_sam=t(xi0_sam + t(Xte %*% xi_sam))
fmean=rowMeans(fhat_sam)
fsort=fhat_sam
for(i in 1:nrow(fhat_sam)){
  fsort[i,]=sort(fhat_sam[i,])
}
f_low=fsort[,ef*0.025]
f_upp=fsort[,ef*0.975]
# save the posterior samples and predictive values
out = list("fhat_sam"=fhat_sam,"fmean"=fmean,"f_low"=f_low,"f_upp"=f_upp)


##### Using bsamGP package (Lenk & Choi, 2017): ######
# MCMC parameters
nblow0 = 1000; # Initialization period for adaptive metropolis
nblow = 1000; # Number of MCMC in transition period
smcmc = 1000; # Number of MCMC for analysis
nskip = 1; # Number of MCMC to skip after nblow
ndisp = 1000; # number of saved draws to be displayed on screen
maxmodmet = 5; # Maximum number of times to modify metropolis
# Prior information
iflagprior = 0; # 1 = Lasso Smoother, 0 = T Smoother
nbasis = 30; # number of cosine basis functions @

# Fit the model
fout = bsar(ytrain ~ fs(xtrain), shape='Increasing',xmin=0, xmax=1,nbasis=nbasis,
            mcmc=list(maxmodmet=maxmodmet,nblow0=nblow0,nblow=nblow,smcmc=smcmc,nskip=nskip,ndisp=ndisp),
            prior=list(iflagprior=iflagprior))
fit = predict(fout, newnp = xtest,type = "mean")
fmean=fit$yhat$mean
f_low=fit$yhat$lower
f_upp=fit$yhat$upper

# save the and predictive values
out = list("fmean"=fmean,"f_low"=f_low,"f_upp"=f_upp)


###############################################################
### Report the error measures and plot of predictive curve ####
###############################################################

### Report the error measures (out-of-sample)
err_out = err.meas.report(data_test,out,reg_i,reg_f,return.mcse=TRUE)  # for bsamGP, return.mcse=FALSE

## total MSPE 
MSPE.total = err_out$rmse
## MSPE over increasing region  
MSPE.inc = err_out$rmse.inc
## MSPE over flat region  
MSPE.flt = err_out$rmse.flt
## coverage 
cvg = err_out$cvg
## effective sample size over 200 test points
ess = err_out$ess 
## standard deviation of the Markov chain
mcse = err_out$mcse


### Report the plot of function predictive curve on test data 
require(ggplot2)
require(grid)

ytr =  yte.tr
x = xtest
y = ytest

fmean = out$fmean
f_low = out$f_low
f_upp = out$f_upp

ub=max(f_low,f_upp,fmean,y)
lb=min(f_low,f_upp,fmean,y)

df = data.frame(x,fmean)
G <- ggplot(df, aes(x=x, y=fmean), color=variable)+
  geom_line(aes(x=x, y=fmean), colour="blue")+
  geom_point(aes(x=x, y=y), colour="green",cex = 0.3)+
  geom_line(aes(x=x, y=ytr), colour="red")+
  geom_ribbon(aes(ymin=f_low, ymax=f_upp), alpha=0.3)
G
G <- G + theme_bw()
Glabs<- G+labs(x = "x", y = "y")
Glabs1<- Glabs + theme(
  axis.title.x = element_text(size=22, face="bold"),
  axis.title.y = element_text(size=22, face="bold"),
  axis.text.x = element_text(size=20),
  axis.text.y = element_text(size=20),
  legend.text = element_text(colour="blue", size=20, face="bold"))
Glabs1<- Glabs1 + ggtitle("")+theme(plot.title = element_text(hjust = 0.5,size=24,face="bold"))
Glabs1

### Add zoomed-in inset plots for the flat region of functions f1,f2,f3, respectively
# ## for f3
# Glabs1
# Gz <- G + coord_cartesian(xlim = c(0.5,1), ylim = c(2.6,3.3), expand = TRUE)+
#   theme(axis.text.x = element_text(size=16),
#         axis.text.y = element_text(size=16))
# 
# Gzlabs<- Gz+labs(x = "", y = "")
# vp <- viewport(width = 0.50, height = 0.32, x = 0.7, y = 0.32)
# print(Gzlabs, vp=vp)


# ## for f2
# Glabs1
# Gz <- G + coord_cartesian(xlim = c(0.6,1), ylim = c(0.5,2), expand = TRUE)+
#   theme(axis.text.x = element_text(size=16),
#         axis.text.y = element_text(size=16))
# Gzlabs<- Gz+labs(x = "", y = "")
# vp <- viewport(width = 0.5, height = 0.3, x = 0.35, y = 0.71)
# print(Gzlabs, vp=vp)

# ## for f1
# Glabs1
# Gz <- G + coord_cartesian(xlim = c(0,0.7), ylim = c(-0.5,0.5), expand = TRUE)+
#   theme(axis.text.x = element_text(size=16),
#         axis.text.y = element_text(size=16))
# Gzlabs<- Gz+labs(x = "", y = "")
# vp <- viewport(width = 0.60, height = 0.36, x = 0.45, y = 0.7)
# print(Gzlabs, vp=vp)




# END 